#ifndef IST_ANIMATED_POLYGON
#define IST_ANIMATED_POLYGON

#include <string>
#include <vector>
#include <map>
#include "ist/ist_math.h"
#include "ist/i3d_polygon_model.h"


namespace ist {
namespace i3d {

class BAnimatedModel : public IAnimatedModel
{
public:
	typedef std::vector<vector3> vertex_cont;
	typedef std::vector<vector2> texcoord_cont;
	typedef std::vector<int>     index_cont;

	class vertex_sequence : public std::map<float, vertex_cont>
	{
	public:
		typedef std::pair<const_iterator, const_iterator> range_t;

		float beginFrame() const {
			if(empty()) return 0;
			return begin()->first;
		}

		float endFrame() const {
			if(empty()) return 0;
			return (--end())->first;
		}

		bool changeFrame(float before, float after) {
			iterator p = find(before);
			if(p==end())
				return false;

			(*this)[after] = p->second;
			erase(p);
			return true;
		}

		range_t range(float f) const {
			return equal_range(f);
		}
		void morph(float t, vertex_cont& t_vertex) const;
	};

	typedef std::map<std::string, vertex_sequence> vertex_clip;
	typedef vertex_clip::const_iterator iterator;


	BAnimatedModel();
	BAnimatedModel(const std::string& filename) { load(filename); }
	~BAnimatedModel();
	void draw(const std::string& clipname, float t) const;
	void draw(iterator p, float t) const;
	void setup();
	bool load(const std::string& filename);
	bool save(const std::string& filename);
	void clear();

	iterator find(const std::string& clipname) { return vtx_clip.find(clipname); }
	iterator begin() { return vtx_clip.begin(); }
	iterator end()   { return vtx_clip.end(); }

	float beginFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.beginFrame();
	}

	float endFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.endFrame();
	}

	bool isExistClip(const std::string& clipname) const { return vtx_clip.find(clipname)!=vtx_clip.end(); }
	const BoundingSphere& getBSphere() const{ return bsphere; }
	const BoundingBox&    getBBox() const   { return bbox; }


	BoundingBox bbox;
	BoundingSphere bsphere;
	vertex_clip   vtx_clip;
	texcoord_cont tex_array;
	index_cont    idx_array;
	unsigned int  num_vertex;
};


class BCachedAnimatedModel : public ICachedAnimatedModel
{
public:
	typedef std::vector<vector3> vertex_cont;
	typedef std::vector<vector3> normal_cont;
	typedef std::vector<vector2> texcoord_cont;
	typedef std::vector<int>     index_cont;

	class vertex_sequence
	{
	public:
		vertex_sequence() : begin_frame(0), end_frame(0) {}
		bool empty() const     { return vertex.empty() || normal.empty(); }
		int beginFrame() const { return begin_frame; }
		int endFrame() const   { return end_frame; }

		int begin_frame;
		int end_frame;
		std::vector<vertex_cont> vertex;
		std::vector<normal_cont> normal;
	};

	typedef std::map<std::string, vertex_sequence> vertex_clip;
	typedef vertex_clip::const_iterator iterator;


	BCachedAnimatedModel();
	BCachedAnimatedModel(const std::string& filename, bool en=true);
	virtual ~BCachedAnimatedModel();
	virtual void draw(const std::string& clipname, int t) const;
	virtual void draw(iterator p, int t) const;

	virtual void setup();
	bool load(const std::string& filename);
	void setVertex(vertex_cont& vtx_array, vertex_cont& vvtx_array);
	void setNormal(vertex_cont& vtx_array, normal_cont& vnml_array);
	void clear();
	void enableNormal(bool en) { enable_normal = en; }

	iterator find(const std::string& clipname) { return vtx_clip.find(clipname); }
	iterator begin() { return vtx_clip.begin(); }
	iterator end()   { return vtx_clip.end(); }

	int beginFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.beginFrame();
	}

	int endFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.endFrame();
	}

	bool isExistClip(const std::string& clipname) const { return vtx_clip.find(clipname)!=vtx_clip.end(); }
	const BoundingSphere& getBSphere() const{ return bsphere; }
	const BoundingBox&    getBBox() const   { return bbox; }


	BoundingBox bbox;
	BoundingSphere bsphere;
	vertex_clip   vtx_clip;
	texcoord_cont tex_array;
	index_cont    idx_array;
	unsigned int  num_vertex;
	float facet;
	bool enable_normal;
};


} //i3d
} //ist

#endif
