#ifndef ISTL_MATRIX_INCLUDE
#define ISTL_MATRIX_INCLUDE


#include "vector.h"

namespace ist {

struct matrix22
{
	union {
		struct {
				float x0, y0;
				float x1, y1;
		};
		float v[4];
		float m[2][2];
	};

	float& operator [](int i) { return v[i]; }

	matrix22 operator * (const matrix22 &mlt) const
	{
		return matrix22(
			v[0]*mlt.v[0] + v[2]*mlt.v[1],
			v[0]*mlt.v[2] + v[2]*mlt.v[3],
			v[1]*mlt.v[0] + v[3]*mlt.v[1],
			v[1]*mlt.v[2] + v[3]*mlt.v[3] );
	}

	void operator *= (const matrix22 &mlt)
	{
		*this=*this*mlt;
	}

	template<typename VecType>
	VecType operator * (const VecType &vec) const
	{
		return VecType(
			v[0]*vec.x + v[2]*vec.y,
			v[1]*vec.x + v[3]*vec.y );
	}

	matrix22()
	{
		identity();
	}

	matrix22(const float *fv):
		x0(fv[0]), y0(fv[1]),
		x1(fv[2]), y1(fv[3]) {}

	matrix22(
		float _0, float _2,
		float _1, float _3 ):
		x0(_0), y0(_1),
		x1(_2), y1(_3) {}

	void identity()
	{
		v[0]=1; v[2]=0;
		v[1]=0; v[3]=1;
	}

	matrix22& invert()
	{
		const int dim=2;
		int i, j, k;
		float p, q;

		for(k=0; k<dim; k++) {
			p=m[k][k];
			m[k][k]=1.0;
			for(j=0; j<dim; j++)
				m[k][j]/=p;
			for(i=0; i<dim; i++) {
				if(i!=k) {
					q=m[i][k];
					m[i][k]=0.0;
					for(j=0; j<dim; j++)
						m[i][j]-=q*m[k][j];
				}
			}
		}

		return *this;
	}

	matrix22& scale(const float *fv)
	{
		scale(fv[0], fv[1]);
		return *this;
	}

	matrix22& scale(float sx, float sy)
	{
		matrix22 sm(sx,0,
					0,sy );
		*this*=sm;
		return *this;
	}


	matrix22& rotate(float ang)
	{
		ang=ang*radian; 
		float cz=cosf(ang), sz=sinf(ang);

		matrix22 rm(cz,-sz,
					sz, cz );
		*this*=rm;
		return *this;
	}
};


struct matrix33
{
	union {
		struct {
				float x0, y0, z0;
				float x1, y1, z1;
				float x2, y2, z2;
		};
		float v[9];
		float m[3][3];
	};

	float& operator [](int i) { return v[i]; }

	matrix33 operator * (const matrix33 &mlt) const
	{
		return matrix33(
			v[0]*mlt.v[0] + v[3]*mlt.v[1] + v[6]*mlt.v[2],
			v[0]*mlt.v[3] + v[3]*mlt.v[4] + v[6]*mlt.v[5],
			v[0]*mlt.v[6] + v[3]*mlt.v[7] + v[6]*mlt.v[8],

			v[1]*mlt.v[0] + v[4]*mlt.v[1] + v[7]*mlt.v[2],
			v[1]*mlt.v[3] + v[4]*mlt.v[4] + v[7]*mlt.v[5],
			v[1]*mlt.v[6] + v[4]*mlt.v[7] + v[7]*mlt.v[8],

			v[2]*mlt.v[0] + v[5]*mlt.v[1] + v[8]*mlt.v[2],
			v[2]*mlt.v[3] + v[5]*mlt.v[4] + v[8]*mlt.v[5],
			v[2]*mlt.v[6] + v[5]*mlt.v[7] + v[8]*mlt.v[8] );
	}

	void operator *= (const matrix33 &mlt)
	{
		*this=*this*mlt;
	}

	template<typename VecType>
	VecType operator * (const VecType &vec) const
	{
		return VecType(
			v[0]*vec.x + v[3]*vec.y + v[6]*vec.z,
			v[1]*vec.x + v[4]*vec.y + v[7]*vec.z,
			v[2]*vec.x + v[5]*vec.y + v[8]*vec.z );
	}

	matrix33()
	{
		identity();
	}

	matrix33(const float *fv):
		x0(fv[0]), y0(fv[1]), z0(fv[2]),
		x1(fv[3]), y1(fv[4]), z1(fv[5]),
		x2(fv[6]), y2(fv[7]), z2(fv[8]) {}

	matrix33(
		float _0, float _3, float _6,
		float _1, float _4, float _7,
		float _2, float _5, float _8 ):
		x0(_0), y0(_1), z0(_2),
		x1(_3), y1(_4), z1(_5),
		x2(_6), y2(_7), z2(_8) {}

	void identity()
	{
		v[0]=1; v[3]=0; v[6]=0;
		v[1]=0; v[4]=1; v[7]=0;
		v[2]=0; v[5]=0; v[8]=1;
	}

	matrix33& invert()
	{
		const int dim=3;
		int i, j, k;
		float p, q;

		for(k=0; k<dim; k++) {
			p=m[k][k];
			m[k][k]=1.0;
			for(j=0; j<dim; j++)
				m[k][j]/=p;
			for(i=0; i<dim; i++) {
				if(i!=k) {
					q=m[i][k];
					m[i][k]=0.0;
					for(j=0; j<dim; j++)
						m[i][j]-=q*m[k][j];
				}
			}
		}

		return *this;
	}

	matrix33& scale(const float *fv)
	{
		scale(fv[0], fv[1], fv[2]);
		return *this;
	}

	matrix33& scale(float sx, float sy, float sz)
	{
		matrix33 S(sx, 0, 0,
					0,sy, 0,
					0, 0,sz );
		*this*=S;
		return *this;
	}


	matrix33& rotate(const float *fv)
	{
		rotate(fv[0], fv[1], fv[2]);
		return *this;
	}

	matrix33& rotate(float p, float h, float r)
	{
		p*=radian;
		h*=radian;
		r*=radian; 
		float sp=sinf(p), cp=cosf(p);
		float sh=sinf(h), ch=cosf(h);
		float sr=sinf(r), cr=cosf(r);

		matrix33 em(cr*ch-sr*sp*sh,-sr*cp, cr*sh+sr*sp*ch,
					sr*ch+cr*sp*sh, cr*cp, sr*sh-cr*sp*ch,
							-cp*sh,    sp,          cp*ch );

		*this=*this*em;
		return *this;
	}


	matrix33& rotateX(float rx)
	{
		rx*=radian;
		float cx=cosf(rx), sx=sinf(rx);

		matrix33 rm(1,  0,  0,
					0, cx,-sx,
					0, sx, cx );
		*this*=rm;
		return *this;
	}

	matrix33& rotateY(float ry)
	{
		ry*=radian;
		float cy=cosf(ry), sy=sinf(ry);

		matrix33 rm( cy, 0, sy,
					  0, 1,  0,
					-sy, 0, cy );
		*this*=rm;
		return *this;
	}

	matrix33& rotateZ(float rz)
	{
		rz*=radian; 
		float cz=cosf(rz), sz=sinf(rz);

		matrix33 rm(cz,-sz, 0,
					sz, cz, 0,
					 0,  0, 1 );
		*this*=rm;
		return *this;
	}


	template<typename VecType>
	matrix33& rotateA(const VecType &ax, float ang)
	{
		ang*=radian;
		const float &x=ax.x, &y=ax.y, &z=ax.z;
		float c=cosf(ang), s=sinf(ang);
		float ic=1-c;

		matrix33 rm(   c+ic*x*x, ic*x*y-z*s, ic*x*z+y*s,
					 ic*x*y+z*s,   c+ic*y*y, ic*y*z-x*s,
					 ic*x*z-y*s, ic*y*z+x*s,   c+ic*z*z );
		*this*=rm;
		return *this;
	}


	template<typename VecType>
	matrix33& aimVector(const VecType &vec)
	{
		{
			static const matrix33 turn90degree = matrix33().rotateY(90);
			static const VecType std_vec(0, 1.0f, 0);
			VecType rotate_axis = (turn90degree*VecType(vec.x, 0, vec.z)).normalize();
			float ang_xz=std_vec.cos(vec);

			rotateA(rotate_axis, ang_xz/radian);
		}
		{
			static const vector2 std_vec(-1.0f, 0);
			vector2 proj_y=vector2(vec.x, vec.z).normalize();
			float ang_y=std_vec.cos(proj_y);
			if(proj_y.y < 0)
				ang_y*=-1;

			rotateY(ang_y/radian-90);
		}

		return *this;
	}

};




struct matrix44
{
	union {
		struct {
				float x0, y0, z0, w0;
				float x1, y1, z1, w1;
				float x2, y2, z2, w2;
				float x3, y3, z3, w3;
		};
		float v[16];
		float m[4][4];
	};


	float& operator [](int i) { return v[i]; }

	matrix44 operator * (const matrix44 &mlt) const
	{
		return matrix44(
		v[0]*mlt.v[ 0] + v[4]*mlt.v[ 1] + v[ 8]*mlt.v[ 2] + v[12]*mlt.v[ 3],
		v[0]*mlt.v[ 4] + v[4]*mlt.v[ 5] + v[ 8]*mlt.v[ 6] + v[12]*mlt.v[ 7],
		v[0]*mlt.v[ 8] + v[4]*mlt.v[ 9] + v[ 8]*mlt.v[10] + v[12]*mlt.v[11],
		v[0]*mlt.v[12] + v[4]*mlt.v[13] + v[ 8]*mlt.v[14] + v[12]*mlt.v[15],

		v[1]*mlt.v[ 0] + v[5]*mlt.v[ 1] + v[ 9]*mlt.v[ 2] + v[13]*mlt.v[ 3],
		v[1]*mlt.v[ 4] + v[5]*mlt.v[ 5] + v[ 9]*mlt.v[ 6] + v[13]*mlt.v[ 7],
		v[1]*mlt.v[ 8] + v[5]*mlt.v[ 9] + v[ 9]*mlt.v[10] + v[13]*mlt.v[11],
		v[1]*mlt.v[12] + v[5]*mlt.v[13] + v[ 9]*mlt.v[14] + v[13]*mlt.v[15],

		v[2]*mlt.v[ 0] + v[6]*mlt.v[ 1] + v[10]*mlt.v[ 2] + v[14]*mlt.v[ 3],
		v[2]*mlt.v[ 4] + v[6]*mlt.v[ 5] + v[10]*mlt.v[ 6] + v[14]*mlt.v[ 7],
		v[2]*mlt.v[ 8] + v[6]*mlt.v[ 9] + v[10]*mlt.v[10] + v[14]*mlt.v[11],
		v[2]*mlt.v[12] + v[6]*mlt.v[13] + v[10]*mlt.v[14] + v[14]*mlt.v[15],

		v[3]*mlt.v[ 0] + v[7]*mlt.v[ 1] + v[11]*mlt.v[ 2] + v[15]*mlt.v[ 3],
		v[3]*mlt.v[ 4] + v[7]*mlt.v[ 5] + v[11]*mlt.v[ 6] + v[15]*mlt.v[ 7],
		v[3]*mlt.v[ 8] + v[7]*mlt.v[ 9] + v[11]*mlt.v[10] + v[15]*mlt.v[11],
		v[3]*mlt.v[12] + v[7]*mlt.v[13] + v[11]*mlt.v[14] + v[15]*mlt.v[15] );
	}

	void operator *= (const matrix44 &mlt)
	{
		*this=*this*mlt;
	}

	template<typename VecType>
	VecType operator * (const VecType &vec) const
	{
		return VecType(
			v[0]*vec.x + v[4]*vec.y + v[ 8]*vec.z + v[12]*vec.w,
			v[1]*vec.x + v[5]*vec.y + v[ 9]*vec.z + v[13]*vec.w,
			v[2]*vec.x + v[6]*vec.y + v[10]*vec.z + v[14]*vec.w,
			v[3]*vec.x + v[7]*vec.y + v[11]*vec.z + v[15]*vec.w );
	}


	matrix44()
	{
		identity();
	}

	matrix44(const float *fv):
		x0(fv[ 0]), y0(fv[ 1]), z0(fv[ 2]), w0(fv[3]), 
		x1(fv[ 4]), y1(fv[ 5]), z1(fv[ 6]), w1(fv[7]), 
		x2(fv[ 8]), y2(fv[ 9]), z2(fv[10]), w2(fv[11]), 
		x3(fv[12]), y3(fv[13]), z3(fv[14]), w3(fv[15]) {}

	matrix44(
		float _00, float _04, float _08, float _12,
		float _01, float _05, float _09, float _13,
		float _02, float _06, float _10, float _14,
		float _03, float _07, float _11, float _15):
		x0(_00), y0(_01), z0(_02), w0(_03),
		x1(_04), y1(_05), z1(_06), w1(_07),
		x2(_08), y2(_09), z2(_10), w2(_11),
		x3(_12), y3(_13), z3(_14), w3(_15) {}


	void identity()
	{
		v[0]=1; v[4]=0; v[ 8]=0; v[12]=0;
		v[1]=0; v[5]=1; v[ 9]=0; v[13]=0;
		v[2]=0; v[6]=0; v[10]=1; v[14]=0;
		v[3]=0; v[7]=0; v[11]=0; v[15]=1;
	}

	matrix44& invert()
	{
		const int dim=4;
		int i, j, k;
		float p, q;

		for(k=0; k<dim; k++) {
			p=m[k][k];
			m[k][k]=1.0;
			for(j=0; j<dim; j++)
				m[k][j]/=p;
			for(i=0; i<dim; i++) {
				if(i!=k) {
					q=m[i][k];
					m[i][k]=0.0;
					for(j=0; j<dim; j++)
						m[i][j]-=q*m[k][j];
				}
			}
		}

		return *this;
	}

	matrix44& translate(const float *fv)
	{
		translate(fv[0], fv[1], fv[2]);
		return *this;
	}

	matrix44& translate(float tx, float ty, float tz)
	{
		matrix44 tm(1, 0, 0, tx,
					0, 1, 0, ty,
					0, 0, 1, tz,
					0, 0, 0,  1 );
		*this*=tm;
		return *this;
	}


	matrix44& scale(const float *fv)
	{
		scale(fv[0], fv[1], fv[2]);
		return *this;
	}

	matrix44& scale(float sx, float sy, float sz)
	{
		matrix44 sm(sx, 0, 0, 0,
					0,sy, 0, 0,
					0, 0,sz, 0,
					0, 0, 0, 1 );
		*this*=sm;
		return *this;
	}


	matrix44& rotate(const float *fv)
	{
		rotate(fv[0], fv[1], fv[2]);
		return *this;
	}

	matrix44& rotate(float p, float h, float r)
	{
		p*=radian;
		h*=radian;
		r*=radian; 
		float sp=sinf(p), cp=cosf(p);
		float sh=sinf(h), ch=cosf(h);
		float sr=sinf(r), cr=cosf(r);

		matrix44 em(cr*ch-sr*sp*sh,-sr*cp, cr*sh+sr*sp*ch, 0,
					sr*ch+cr*sp*sh, cr*cp, sr*sh-cr*sp*ch, 0,
							-cp*sh,    sp,          cp*ch, 0,
								 0,     0,              0, 1 );

		*this=*this*em;
		return *this;
	}


	matrix44& rotateX(float rx)
	{
		rx*=radian;
		float cx=cosf(rx), sx=sinf(rx);

		matrix44 rm(1,  0,  0, 0,
					0, cx,-sx, 0,
					0, sx, cx, 0,
					0,  0,  0, 1 );
		*this*=rm;
		return *this;
	}

	matrix44& rotateY(float ry)
	{
		ry*=radian;
		float cy=cosf(ry), sy=sinf(ry);

		matrix44 rm( cy, 0, sy, 0,
					  0, 1,  0, 0,
					-sy, 0, cy, 0,
					  0, 0,  0, 1 );
		*this*=rm;
		return *this;
	}

	matrix44& rotateZ(float rz)
	{
		rz*=radian; 
		float cz=cosf(rz), sz=sinf(rz);

		matrix44 rm(cz,-sz, 0, 0,
					sz, cz, 0, 0,
					 0,  0, 1, 0,
					 0,  0, 0, 1 );
		*this*=rm;
		return *this;
	}


	template<typename VecType>
	matrix44& rotateA(const VecType &ax, float ang)
	{
		ang*=radian;
		const float &x=ax.x, &y=ax.y, &z=ax.z;
		float c=cosf(ang), s=sinf(ang);
		float ic=1-c;

		matrix44 rm(   c+ic*x*x, ic*x*y-z*s, ic*x*z+y*s, 0,
					 ic*x*y+z*s,   c+ic*y*y, ic*y*z-x*s, 0,
					 ic*x*z-y*s, ic*y*z+x*s,   c+ic*z*z, 0,
							  0,          0,          0, 1 );
		*this*=rm;
		return *this;
	}


	template<typename VecType>
	matrix44& aimVector(const VecType &vec)
	{
		{
			static const matrix33 turn90degree = matrix33().rotateY(90);
			static const VecType std_vec(0, 1.0f, 0);
			VecType rotate_axis = (turn90degree*VecType(vec.x, 0, vec.z)).normalize();
			float ang_xz=std_vec.cos(vec);

			rotateA(rotate_axis, ang_xz/radian);
		}
		{
			static const vector2 std_vec(-1.0f, 0);
			vector2 proj_y=vector2(vec.x, vec.z).normalize();
			float ang_y=std_vec.cos(proj_y);
			if(proj_y.y < 0)
				ang_y*=-1;

			rotateY(ang_y/radian-90.0f);
		}

		return *this;
	}

};


} // ist

#endif

