#include "socket.h"



namespace ist {

#ifdef _USE_WINSOCK

class WinSockInitializer
{
public:
	WinSockInitializer()
	{
		WSADATA info;
		int res=WSAStartup(0x0002, &info);
		(res==SOCKET_ERROR) ? is_normal=false : is_normal=true;
	}

	~WinSockInitializer()
	{
		if(is_normal)
			WSACleanup();
	}

	WinSockInitializer& operator =(const WinSockInitializer&);

	bool is_normal;
};
	
#endif

static socket_t create_socket(int type) {
#ifdef _USE_WINSOCK
	static WinSockInitializer wsi;
#endif
	return ::socket(AF_INET, type, 0);
}



Socket::Socket(socket_t s) : _sock(s) {}
Socket::Socket(const char *host, unsigned short port) : _sock(-1) { open(host, port); }

Socket::~Socket()
{
	if(is_opened()) {
#ifdef _USE_WINSOCK
		::closesocket(_sock);
#else
		::close(_sock);
#endif
	}
}

bool Socket::open(const char *host, unsigned short port) {
	return open(SOCK_STREAM, host, port);
}

bool Socket::open(int type, const char *host, unsigned short port) {
	if(is_opened()) close();

	_sock= create_socket(type);
	if (_sock==-1) return false;

	sockaddr_in addr;
	addr.sin_family=AF_INET;
	addr.sin_port=htons(port);

	unsigned long uaddr=::inet_addr(host);
	if(uaddr==INADDR_NONE){
		hostent *ent=::gethostbyname(host);
		if (!ent || !ent->h_addr_list[0]){
			close();
			return false;
		}
		uaddr=*reinterpret_cast<unsigned long*>(ent->h_addr_list[0]);
	}

#ifdef _USE_WINSOCK
	addr.sin_addr.S_un.S_addr = uaddr;
#else
	addr.sin_addr.s_addr = uaddr;
#endif

	if(::connect(_sock, reinterpret_cast<sockaddr*>(&addr), sizeof(sockaddr_in))==-1)
		close();

	return is_opened();
}

void Socket::close() {
	if (_sock==-1) return;
#ifdef _USE_WINSOCK
	::closesocket(_sock);
#else
	::close(_sock);
#endif
	_sock= -1;
}

bool Socket::write(const char *str, int length) {
	if (!is_opened()) return false;
	if( send(_sock, str, length, 0)!=-1 )
		return true;
	else
		return false;
}

int Socket::read(char *c, int length) {
	if (!is_opened()) return -1;
	return recv(_sock, c, length, 0);
}

bool Socket::is_opened() const { return _sock!=-1; }

bool Socket::operator!() const { return !is_opened(); }





ServerSocket::ServerSocket()
	: _sock(create_socket(SOCK_STREAM)), _is_listened(false) {}

ServerSocket::ServerSocket(int port, int backlog)
	: _sock(create_socket(SOCK_STREAM)), _is_listened(false) {
	bind(port, backlog);  
}

ServerSocket::ServerSocket(int type, int port, int backlog)
	: _sock(create_socket(type)), _is_listened(false) {
	bind(port, backlog);  
}

ServerSocket::~ServerSocket() {
	close();
}

bool ServerSocket::bind(int port, int backlog) {
	if (_is_listened || _sock==-1) return false;

	sockaddr_in addr;
	addr.sin_family= AF_INET;
	addr.sin_port= htons(port);
#ifdef _USE_WINSOCK
	addr.sin_addr.S_un.S_addr= INADDR_ANY;
#else
	addr.sin_addr.s_addr= INADDR_ANY;
#endif

	int on=1;
	setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, reinterpret_cast<char*>(&on), sizeof(int));

	if(::bind(_sock, reinterpret_cast<sockaddr*>(&addr), sizeof(addr))==-1)
		return false;

	if(::listen(_sock, backlog)==-1) {
		close();
		return false;
	}

	_is_listened= true;
	return true;
}


void ServerSocket::close() {
	if (_sock==-1) return;

#ifdef _USE_WINSOCK
	::closesocket(_sock);
#else
	::close(_sock);
#endif

	_sock= -1;
}

socket_t ServerSocket::accept() {
	return ::accept(_sock, 0, 0);
}

bool ServerSocket::operator!() const { return !_is_listened; }


} //ist




